-- |
-- Module       : Data.ByteString.Base64
-- Copyright 	: (c) 2019 Emily Pillmore
-- License	: BSD-style
--
-- Maintainer	: Emily Pillmore <emilypi@cohomolo.gy>
-- Stability	: Experimental
-- Portability	: portable
--
-- This module contains the combinators implementing the
-- RFC 4648 specification for the Base64 encoding including
-- unpadded and lenient variants
--
module Data.ByteString.Base64
( encodeBase64
, decodeBase64
, encodeBase64Unpadded
, decodeBase64Unpadded
) where


import Data.ByteString (ByteString)
import Data.ByteString.Base64.Internal
import Data.Text (Text)


-- | Encode a 'ByteString' in base64 with padding.
--
-- See: RFC-4648 section 4
--
encodeBase64 :: ByteString -> ByteString
encodeBase64 = encodeB64Padded base64Table
{-# INLINE encodeBase64 #-}

-- | Decode a padded base64-encoded 'ByteString'
--
-- See: RFC-4648 section 4
--
decodeBase64 :: ByteString -> Either Text ByteString
decodeBase64 = decodeB64 decodeB64Table
{-# INLINE decodeBase64 #-}

-- | Encode a 'ByteString' in base64 without padding.
--
-- Note: in some circumstances, the use of padding ("=") in base-encoded data
-- is not required or used. This is not one of them. If you are absolutely sure
-- the length of your bytestring is divisible by 3, this function will be the same
-- as 'encodeBase64' with padding, however, if not, you may see garbage appended to
-- your bytestring in the form of "\NUL".
--
-- Only call unpadded variants when you can make assumptions about the length of
-- your input data.
--
-- See: RFC-4648 section 3.2
--
encodeBase64Unpadded :: ByteString -> ByteString
encodeBase64Unpadded = encodeB64Unpadded base64Table
{-# INLINE encodeBase64Unpadded #-}

-- | Decode an unpadded base64-encoded 'ByteString'
--
-- See: RFC-4648 section 3.2
--
decodeBase64Unpadded :: ByteString -> Either Text ByteString
decodeBase64Unpadded = decodeB64 decodeB64Table
{-# INLINE decodeBase64Unpadded #-}
